#!/usr/bin/env python

"""
weatherquery.py - Demonstrates how to use XQuery from Python.
                  See http://www.yr.no/verdata/1.3321307 for the terms of use
                  of the yr.no weather service.

This file is part of the PyQt for Desktop and Embedded Devices package.

Copyright (C) 2009 David Boddie <david.boddie@nokia.com>
Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).
All rights reserved.

Contact: Nokia Corporation (qt-info@nokia.com)

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names
   of its contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
"""

import os, sys
from PyQt4.QtCore import QBuffer, QByteArray, QDate, QDateTime, \
                         QFile, QSize, QStringList, Qt, QVariant, SIGNAL
from PyQt4.QtGui import QApplication, QCheckBox, QComboBox, QFormLayout, \
                        QIcon, QLabel, QTableWidget, QTableWidgetItem, \
                        QVBoxLayout, QWidget
from PyQt4.QtXmlPatterns import *

query_string = \
"""declare variable $url external;

(
string(doc($url)//sun/@rise),
string(doc($url)//sun/@set),

for $time in doc($url)//tabular/time
order by $time/@from
return (string($time/@from),
        string($time/symbol/@name), string($time/symbol/@number),
        string($time/temperature/@value), string($time/temperature/@unit))
)
"""

This_Dir = os.path.split(__file__)[0]

symbols = {"1": "01d.png",
           "2": "02d.png",
           "3": "03d.png",
           "4": "04.png",
           "5": "05d.png",
           "6": "06d.png",
           "7": "07d.png",
           "8": "08d.png",
           "9": "09.png",
           "10": "10.png",
           "11": "11.png",
           "12": "12.png",
           "13": "13.png",
           "14": "14.png",
           "15": "15.png"}

class WeatherWidget(QWidget):

    def __init__(self, parent = None):
    
        QWidget.__init__(self, parent)
        
        self.results = QStringList()
        
        placeCombo = QComboBox()
        placeCombo.addItem("Italia/Toscana/Firenze")
        placeCombo.addItem("Norway/Oslo/Oslo/Oslo")
        placeCombo.addItem("Tyskland/Berlin/Berlin")
        
        self.todayCheckBox = QCheckBox(self.tr("&Only today"))
        
        self.forecastTable = QTableWidget()
        self.forecastTable.setColumnCount(3)
        self.forecastTable.horizontalHeader().hide()
        self.forecastTable.verticalHeader().hide()
        self.forecastTable.horizontalHeader().setStretchLastSection(True)
        self.forecastTable.setIconSize(QSize(50, 50))
        
        creditLabel = QLabel(self.tr("Weather forecast from yr.no, delivered "
                                     "by the Norwegian Meteorological "
                                     "Institute and the NRK"))
        creditLabel.setWordWrap(True)
        
        self.connect(placeCombo, SIGNAL("currentIndexChanged(QString)"),
                     self.fetchForecast)
        self.connect(self.todayCheckBox, SIGNAL("toggled(bool)"),
                     self.updateTable)
        
        layout = QFormLayout()
        layout.addRow(self.tr("&Place:"), placeCombo)
        layout.addRow(self.tr("&Forecast:"), self.todayCheckBox)
        
        mainLayout = QVBoxLayout(self)
        mainLayout.addLayout(layout)
        mainLayout.addWidget(self.forecastTable)
        mainLayout.addWidget(creditLabel)
        
        self.setWindowTitle(self.tr("Weather Forecast"))
        self.fetchForecast("Italia/Toscana/Firenze")
    
    def fetchForecast(self, place):
    
        url = "http://www.yr.no/place/" + place + "/forecast.xml"
        #url = "file:///home/david/Private/Python/PyQt4/Documents/PyCon Italia 2009/Examples/XMLPatterns/forecast.xml"
        
        b = QBuffer()
        b.setData(query_string)
        b.open(QBuffer.ReadOnly)
        
        query = QXmlQuery()
        query.bindVariable("url", QXmlItem(QVariant(url)))
        query.setQuery(b)
        b.close()
        
        if query.isValid():
        
            self.results.clear()
            query.evaluateTo(self.results)
            self.updateTable()
    
    def updateTable(self):
    
        if len(self.results):
        
            self.forecastTable.clear()
            self.forecastTable.setRowCount(0)
            
            sunrise = QDateTime.fromString(self.results[0], Qt.ISODate)
            sunset = QDateTime.fromString(self.results[1], Qt.ISODate)
            
            i = 2
            while i < len(self.results):
            
                datetime = QDateTime.fromString(self.results[i], Qt.ISODate)
                onlyToday = self.todayCheckBox.checkState() == Qt.Checked
                
                if not onlyToday or datetime.date() == QDate.currentDate():
                
                    description = self.results[i + 1]
                    identifier = self.results[i + 2]
                    temperature = self.results[i + 3]
                    unit = self.results[i + 4]
                    if unit == u"celcius":
                        unit = u"\u00b0C"
                    
                    row = self.forecastTable.rowCount()
                    self.forecastTable.insertRow(row)
                    
                    if onlyToday:
                        timeItem = QTableWidgetItem(datetime.time().toString())
                    else:
                        timeItem = QTableWidgetItem(datetime.toString("dddd\nhh:mm"))
                    timeItem.setFlags(Qt.ItemIsEnabled)
                    self.forecastTable.setItem(row, 0, timeItem)
                    
                    try:
                        icon = QIcon(os.path.join(This_Dir, "images", symbols[str(identifier)]))
                        symbolItem = QTableWidgetItem(icon, description)
                    except KeyError:
                        symbolItem = QTableWidgetItem(description)
                    
                    symbolItem.setFlags(Qt.ItemIsEnabled)
                    
                    self.forecastTable.setItem(row, 1, symbolItem)
                    
                    temperatureItem = QTableWidgetItem(temperature + u" " + unit)
                    temperatureItem.setFlags(Qt.ItemIsEnabled)
                    self.forecastTable.setItem(row, 2, temperatureItem)
                
                i += 5
            
            self.forecastTable.resizeColumnsToContents()
            self.forecastTable.resizeRowsToContents()


if __name__ == "__main__":

    app = QApplication(sys.argv)
    
    window = WeatherWidget()
    window.show()
    
    sys.exit(app.exec_())
