"""
items.py - Definitions of graphics items used in the presentation.
This file is part of the PyQt for Desktop and Embedded Devices package.

Copyright (C) 2009 David Boddie <david.boddie@nokia.com>
Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).
All rights reserved.

Contact: Nokia Corporation (qt-info@nokia.com)

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names
   of its contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
"""

import math
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from constants import *
import highlighter


class PresentationItem:

    def animate(self, value):
    
        pass
    
    def transit(self, value, view):
    
        pass
    
    def anchorAt(self, position, alignment):
    
        size = self.transform().map(QPointF(self.size().width(), self.size().height()))
        width, height = size.x(), size.y()
        
        dx = 0
        if alignment & Qt.AlignRight:
            dx = -width
        elif alignment & Qt.AlignHCenter:
            dx = -width/2.0
        
        dy = 0
        if alignment & Qt.AlignBottom:
            dy = -height
        elif alignment & Qt.AlignVCenter:
            dy = -height/2.0
        
        self.setPos(position + QPointF(dx, dy))
    
    def size(self):
        return self.boundingRect().size()
    
    def rect(self):
        # Note: Subclasses of QGraphicsRectItem will override this implementation.
        return QRectF(self.pos() + self.boundingRect().topLeft(),
                      self.size())
    
    def center(self):
        return self.pos() + self.boundingRect().center()
    
    def top(self):
        return self.pos() + QPointF(self.boundingRect().center().x(),
                                    self.boundingRect().top())
    def topLeft(self):
        return self.pos() + self.boundingRect().topLeft()
    
    def topRight(self):
        return self.pos() + self.boundingRect().topRight()
    
    def bottom(self):
        return self.pos() + QPointF(self.boundingRect().center().x(),
                                    self.boundingRect().bottom())
    def bottomLeft(self):
        return self.pos() + self.boundingRect().bottomLeft()
    
    def bottomRight(self):
        return self.pos() + self.boundingRect().bottomRight()
    
    def left(self):
        return self.pos() + QPointF(self.boundingRect().left(),
                                    self.boundingRect().center().y())
    def right(self):
        return self.pos() + QPointF(self.boundingRect().right(),
                                    self.boundingRect().center().y())


# Mix-ins

class Fading:

    def __init__(self, pairs):
    
        self.start = pairs[0][0]
        self.finish = pairs[-1][0]
        self.setOpacity(0.0)
        
        self.intervals = []
        
        if pairs[0][0] > -1.0:
            next, next_opacity = pairs[0]
            self.intervals.append((-1.0, next, next_opacity, 0.0))
        
        for i in range(len(pairs) - 1):
        
            this, this_opacity = pairs[i]
            next, next_opacity = pairs[i + 1]
            self.intervals.append((this, next, this_opacity, float(next_opacity - this_opacity)/(next - this)))
        
        if pairs[-1][0] < 1.0:
            this, this_opacity = pairs[-1]
            self.intervals.append((this, 1.0, this_opacity, 0.0))
    
    def transit(self, value, view):
    
        for this, next, opacity, gradient in self.intervals:
        
            if this <= value <= next:
            
                self.setOpacity(opacity + (value - this) * gradient)
                break


class Recoloring:

    def __init__(self, pairs):
    
        self.start = pairs[0][0]
        self.finish = pairs[-1][0]
        self.setColor(Transparent_Color)
        
        self.intervals = []
        
        if pairs[0][0] > -1.0:
            next, next_color = pairs[0]
            self.intervals.append((-1.0, next, next_color, (0.0, 0.0, 0.0, 0.0)))
        
        for i in range(len(pairs) - 1):
        
            this, this_color = pairs[i]
            next, next_color = pairs[i + 1]
            dr = (next_color.red() - this_color.red())/(next - this)
            dg = (next_color.green() - this_color.green())/(next - this)
            db = (next_color.blue() - this_color.blue())/(next - this)
            da = (next_color.alpha() - this_color.alpha())/(next - this)
            self.intervals.append((this, next, this_color, (dr, dg, db, da)))
        
        if pairs[-1][0] < 1.0:
            this, this_color = pairs[-1]
            self.intervals.append((this, 1.0, this_color, (0.0, 0.0, 0.0, 0.0)))
    
    def transit(self, value, view):
    
        for this, next, color, (dr, dg, db, da) in self.intervals:
        
            if this <= value < next:
            
                r = color.red() + (value - this) * dr
                g = color.green() + (value - this) * dg
                b = color.blue() + (value - this) * db
                a = color.alpha() + (value - this) * da
                
                self.setBrush(QColor(r, g, b, a))
                break


class Outlined:

    def animate(self, value):
    
        phase = 2 * value * math.pi
        intensity = 128 + math.cos(phase) * 127
        color = self.color
        color.setAlpha(intensity)
        self.setOutline(QPen(color, self.thickness))


# Items

class Slide(QGraphicsRectItem, PresentationItem):

    def __init__(self, rect, scene, printable = True):
    
        QGraphicsRectItem.__init__(self, rect)
        scene.addItem(self)
        self.setPen(QPen(Qt.NoPen))
        
        self.printable = printable
        self.items = []
        self.previous = None
        self.next = None
        self.next_rect = rect
        
        self.timeLine = QTimeLine(1000)
        self.timeLine.setCurveShape(QTimeLine.LinearCurve)
        self.timeLine.setFrameRange(0, 100)
        self.timeLine.setUpdateInterval(25)
        QObject.connect(self.timeLine, SIGNAL("valueChanged(qreal)"), self.animate)
        QObject.connect(self.timeLine, SIGNAL("finished()"), self.timeLine.start)
    
    def addSlide(self, slide):
    
        self.next = slide
        slide.previous = self
        self.next_rect = slide.rect()
        
        self.dx = self.next_rect.x() - self.rect().x()
        self.dy = self.next_rect.y() - self.rect().y()
        self.dw = self.next_rect.width() - self.rect().width()
        self.dh = self.next_rect.height() - self.rect().height()
        
        return self
    
    def transit(self, value, view):
    
        if self.next:
        
            v = 0.5 * (1 - math.cos((value**0.65) * math.pi))
            t_old = view.transform()
            
            if value == 0:
                w = self.rect().width()
                h = self.rect().height()
                x = self.rect().x()
                y = self.rect().y()
            elif value == 1:
                w = self.next_rect.width()
                h = self.next_rect.height()
                x = self.next_rect.x()
                y = self.next_rect.y()
            else:
                x = self.rect().x() + self.dx * v
                y = self.rect().y() + self.dy * v
                w = self.rect().width() + self.dw * v
                h = self.rect().height() + self.dh * v
            
            t = QTransform(1, 0, 0, 0, 1, 0, -x, -y, 1)
            if t != t_old:
                view.setTransform(t, False)
            else:
                view.viewport().update()
        
        for item in self.items:
            item.transit(value, view)
        
        if self.next:
            for item in self.next.items:
                item.transit(value - 1, view)
    
    def startAnimations(self):
    
        self.timeLine.setCurrentTime(0)
        self.timeLine.start()
    
    def stopAnimations(self):
    
        self.timeLine.stop()
    
    def animate(self, value):
    
        for item in self.items:
            item.animate(value)


class TargetItem(PresentationItem):

    def __init__(self, initial, final):
    
        self.initial = initial
        self.final = final
    
    def transit(self, value, view):
    
        difference = self.final - self.initial
        pos = self.initial + QPointF(difference.x() * value, difference.y() * value)
        view.centerOn(pos)

    
class LabelItem(QGraphicsSimpleTextItem, PresentationItem):

    def __init__(self, text, scene):
    
        QGraphicsSimpleTextItem.__init__(self, text)
        scene.addItem(self)
        self.setZValue(1.0)
    
    def anchorAt(self, position, alignment):
    
        PresentationItem.anchorAt(self, position, alignment)
        if alignment & Qt.AlignVCenter:
            metrics = QFontMetricsF(self.font())
            self.translate(0, -metrics.descent() / 2.0)
    
    def setColor(self, color):
        self.setBrush(color)

class HTMLItem(QGraphicsTextItem, PresentationItem):

    def __init__(self, text, scene):
    
        QGraphicsTextItem.__init__(self)
        scene.addItem(self)
        self.setHtml(text)
        self.setZValue(1.0)
    
    def setColor(self, color):
        self.setDefaultTextColor(color)

class FadingLabel(Fading, LabelItem):

    def __init__(self, text, pairs, scene):
    
        LabelItem.__init__(self, text, scene)
        Fading.__init__(self, pairs)


class RecoloringLabel(Recoloring, LabelItem):

    def __init__(self, text, pairs, scene):
    
        LabelItem.__init__(self, text, scene)
        Recoloring.__init__(self, pairs)


class RectangleItem(PresentationItem, QGraphicsRectItem):

    def __init__(self, rect, scene):
    
        QGraphicsRectItem.__init__(self, rect)
        scene.addItem(self)
        self.setPen(QPen(Qt.NoPen))
    
    def setColor(self, color):
        self.setBrush(color)
    
    def setOutline(self, color):
        self.setPen(color)

class FadingBoxItem(Fading, RectangleItem):

    def __init__(self, rect, pairs, scene):
    
        RectangleItem.__init__(self, rect, scene)
        Fading.__init__(self, pairs)

class OutlinedBoxItem(Outlined, FadingBoxItem):

    def __init__(self, rect, pairs, scene):
    
        FadingBoxItem.__init__(self, rect, pairs, scene)
    
    def setColor(self, color, thickness = 4):
    
        self.color = color
        self.thickness = thickness

class PathItem(Fading, QGraphicsPathItem, PresentationItem):

    def __init__(self, path, pairs, scene):
    
        QGraphicsPathItem.__init__(self, path)
        Fading.__init__(self, pairs)
        scene.addItem(self)
        self.setPen(QPen(Qt.NoPen))


class WidgetItem(Fading, QGraphicsProxyWidget, PresentationItem):

    def __init__(self, widget, pairs, scene):
    
        QGraphicsProxyWidget.__init__(self)
        Fading.__init__(self, pairs)
        self.setWidget(widget)
        scene.addItem(self)


class SourceItem(QGraphicsTextItem, PresentationItem):

    def __init__(self, text, font, scene):
    
        QGraphicsTextItem.__init__(self, text)
        scene.addItem(self)


class FadingSource(Fading, SourceItem):

    def __init__(self, text, font, pairs, scene):
    
        SourceItem.__init__(self, text, font, scene)
        Fading.__init__(self, pairs)


class PythonSourceItem(SourceItem):

    def __init__(self, text, font, scene):
    
        SourceItem.__init__(self, text, font, scene)
        
        char_format = QTextCharFormat()
        char_format.setFont(font)
        
        self.highlighter = highlighter.PythonHighlighter(self.document(), char_format)


class FadingPythonSource(Fading, PythonSourceItem):

    def __init__(self, text, font, pairs, scene):
    
        PythonSourceItem.__init__(self, text, font, scene)
        Fading.__init__(self, pairs)


class PixmapItem(QGraphicsPixmapItem, PresentationItem):

    def __init__(self, pixmap, scene):
    
        QGraphicsPixmapItem.__init__(self, pixmap)
        scene.addItem(self)


class FadingPixmap(Fading, QGraphicsPixmapItem, PresentationItem):

    def __init__(self, pixmap, pairs, scene):
    
        QGraphicsPixmapItem.__init__(self, pixmap)
        Fading.__init__(self, pairs)
        scene.addItem(self)


class Edge(QGraphicsItem):
    Pi = math.pi
    TwoPi = 2.0 * Pi

    def __init__(self, sourceNode, destNode, parent):
    
        QGraphicsItem.__init__(self, parent)
        
        self.arrowSize = 10.0
        self.sourcePoint = QPointF()
        self.destPoint = QPointF()

        self.source = sourceNode
        self.dest = destNode
        self.source.addEdge(self)
        self.dest.addEdge(self)
        self.adjust()

    def adjust(self):
        if not self.source or not self.dest:
            return

        line = QLineF(self.mapFromItem(self.source, 0, 0),
                self.mapFromItem(self.dest, 0, 0))
        length = line.length()

        if length == 0.0:
            return

        edgeOffset = QPointF((line.dx() * 10) / length,
                (line.dy() * 10) / length)

        self.prepareGeometryChange()
        self.sourcePoint = line.p1() + edgeOffset
        self.destPoint = line.p2() - edgeOffset

    def boundingRect(self):
        if not self.source or not self.dest:
            return QRectF()

        penWidth = 1
        extra = (penWidth + self.arrowSize) / 2.0

        return QRectF(self.sourcePoint,
                      QSizeF(self.destPoint.x() - self.sourcePoint.x(),
                             self.destPoint.y() - self.sourcePoint.y())).normalized().adjusted(-extra, -extra, extra, extra)

    def paint(self, painter, option, widget):
        if not self.source or not self.dest:
            return

        # Draw the line itself.
        line = QLineF(self.sourcePoint, self.destPoint)

        if line.length() == 0.0:
            return

        painter.setPen(QPen(Qt.black, 1, Qt.SolidLine,
                Qt.RoundCap, Qt.RoundJoin))
        painter.drawLine(line)

        # Draw the arrows if there's enough room.
        angle = math.acos(line.dx() / line.length())
        if line.dy() >= 0:
            angle = Edge.TwoPi - angle

        sourceArrowP1 = self.sourcePoint + QPointF(math.sin(angle + Edge.Pi / 3) * self.arrowSize,
                                                          math.cos(angle + Edge.Pi / 3) * self.arrowSize)
        sourceArrowP2 = self.sourcePoint + QPointF(math.sin(angle + Edge.Pi - Edge.Pi / 3) * self.arrowSize,
                                                          math.cos(angle + Edge.Pi - Edge.Pi / 3) * self.arrowSize);   
        destArrowP1 = self.destPoint + QPointF(math.sin(angle - Edge.Pi / 3) * self.arrowSize,
                                                      math.cos(angle - Edge.Pi / 3) * self.arrowSize)
        destArrowP2 = self.destPoint + QPointF(math.sin(angle - Edge.Pi + Edge.Pi / 3) * self.arrowSize,
                                                      math.cos(angle - Edge.Pi + Edge.Pi / 3) * self.arrowSize)

        painter.setBrush(Qt.black)
        painter.drawPolygon(QPolygonF([line.p1(), sourceArrowP1, sourceArrowP2]))
        painter.drawPolygon(QPolygonF([line.p2(), destArrowP1, destArrowP2]))


class Node(QGraphicsItem):

    def __init__(self, parent):
    
        QGraphicsItem.__init__(self, parent)
        
        self.edgeList = []
        self.newPos = QPointF()
        self.setFlag(QGraphicsItem.ItemIsMovable)

    def addEdge(self, edge):
        self.edgeList.append(edge)
        edge.adjust()

    def calculateForces(self):
        if not self.scene() or self.scene().mouseGrabberItem() is self:
            self.newPos = self.pos()
            return
    
        # Sum up all forces pushing this item away.
        xvel = 0.0
        yvel = 0.0
        for item in self.scene().items():
            if not isinstance(item, Node):
                continue

            line = QLineF(self.mapFromItem(item, 0, 0),
                    QPointF(0, 0))
            dx = line.dx()
            dy = line.dy()
            l = 2.0 * (dx * dx + dy * dy)
            if l > 0:
                xvel += (dx * 150.0) / l
                yvel += (dy * 150.0) / l

        # Now subtract all forces pulling items together.
        weight = (len(self.edgeList) + 1) * 10.0
        for edge in self.edgeList:
            if edge.source is self:
                pos = self.mapFromItem(edge.dest, 0, 0)
            else:
                pos = self.mapFromItem(edge.source, 0, 0)
            xvel += pos.x() / weight
            yvel += pos.y() / weight
    
        if qAbs(xvel) < 0.1 and qAbs(yvel) < 0.1:
            xvel = yvel = 0.0

        sceneRect = self.scene().sceneRect()
        self.newPos = self.pos() + QPointF(xvel, yvel)
        self.newPos.setX(min(max(self.newPos.x(), sceneRect.left() + 10), sceneRect.right() - 10))
        self.newPos.setY(min(max(self.newPos.y(), sceneRect.top() + 10), sceneRect.bottom() - 10))

    def advance(self):
        if self.newPos == self.pos():
            return False

        self.setPos(self.newPos)
        return True

    def boundingRect(self):
        adjust = 2.0
        return QRectF(-10 - adjust, -10 - adjust, 23 + adjust,
                23 + adjust)

    def shape(self):
        path = QPainterPath()
        path.addEllipse(-10, -10, 20, 20)
        return path

    def paint(self, painter, option, widget):
        painter.setPen(Qt.NoPen)
        painter.setBrush(Qt.darkGray)
        painter.drawEllipse(-7, -7, 20, 20)

        gradient = QRadialGradient(-3, -3, 10)
        if option.state & QStyle.State_Sunken:
            gradient.setCenter(3, 3)
            gradient.setFocalPoint(3, 3)
            gradient.setColorAt(1, QColor(Qt.yellow).light(120))
            gradient.setColorAt(0, QColor(Qt.darkYellow).light(120))
        else:
            gradient.setColorAt(0, Qt.yellow)
            gradient.setColorAt(1, Qt.darkYellow)

        painter.setBrush(QBrush(gradient))
        painter.setPen(QPen(Qt.black, 0))
        painter.drawEllipse(-10, -10, 20, 20)

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            for edge in self.edgeList:
                edge.adjust()

        return QGraphicsItem.itemChange(self, change, value)


class NodesContainer(Fading, QGraphicsRectItem, PresentationItem):

    def __init__(self, pairs, scene):
    
        QGraphicsRectItem.__init__(self)
        Fading.__init__(self, pairs)
        scene.addItem(self)
    
    def animate(self, value):
    
        for node in self.nodes:
            node.calculateForces()

        for node in self.nodes:
            node.advance()
    
    def anchorAt(self, position, alignment):
    
        PresentationItem.anchorAt(self, position, alignment)
        
        w = self.rect().width()
        h = self.rect().height()
        
        node1 = Node(self)
        node2 = Node(self)
        node3 = Node(self)
        node4 = Node(self)
        self.centerNode = Node(self)
        node6 = Node(self)
        node7 = Node(self)
        node8 = Node(self)
        node9 = Node(self)
        
        self.nodes = [node1, node2, node3, node4, self.centerNode, node6,
                      node7, node8, node9]
        
        self.nodes[0].setPos(0.25 * w, 0.25 * h)
        self.nodes[1].setPos(0.5 * w, 0.25 * h)
        self.nodes[2].setPos(0.75 * w, 0.25 * h)
        self.nodes[3].setPos(0.25 * w, 0.5 * h)
        self.centerNode.setPos(0.5 * w, 0.5 * h)
        self.nodes[5].setPos(0.75 * w, 0.5 * h)
        self.nodes[6].setPos(0.25 * w, 0.75 * h)
        self.nodes[7].setPos(0.5 * w, 0.75 * h)
        self.nodes[8].setPos(0.75 * w, 0.75 * h)
        
        Edge(node1, node2, self)
        Edge(node2, node3, self)
        Edge(node2, self.centerNode, self)
        Edge(node3, node6, self)
        Edge(node4, node1, self)
        Edge(node4, self.centerNode, self)
        Edge(self.centerNode, node6, self)
        Edge(self.centerNode, node8, self)
        Edge(node6, node9, self)
        Edge(node7, node4, self)
        Edge(node8, node7, self)
        Edge(node9, node8, self)
