#!/usr/bin/env python

"""
main.pyw - The main Python program for the presentation.

Original copyright (C) 2009 David Boddie <david.boddie@nokia.com>
This file is hereby placed into the public domain.
"""

import sys, types
from PyQt4.QtCore import *
from PyQt4.QtGui import *
import constants
import slides

class View(QGraphicsView):

    transition_duration = 1000
    ignoreKeys = (Qt.Key_Up, Qt.Key_Down, Qt.Key_PageUp, Qt.Key_PageDown)
    
    def __init__(self, scene, parent = None):
    
        QGraphicsView.__init__(self, scene, parent)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setTransformationAnchor(self.NoAnchor)
        self.setRenderHint(QPainter.Antialiasing)
        
        self.timeLine = QTimeLine(self.transition_duration)
        self.timeLine.setCurveShape(QTimeLine.LinearCurve)
        self.timeLine.setFrameRange(0, 100)
        self.timeLine.setUpdateInterval(20)
        self.connect(self.timeLine, SIGNAL("valueChanged(qreal)"), self.updateTransition)
        self.connect(self.timeLine, SIGNAL("finished()"), self.updateSlide)
        
        self.slides = None
        self.firstSlide = None
        self.currentSlide = None
        
        exitAction = QAction(self.tr("E&xit"), self)
        exitAction.setShortcut(QKeySequence(self.tr("Ctrl+Q")))
        self.connect(exitAction, SIGNAL("triggered()"), qApp.quit)
        self.addAction(exitAction)
        
        forwardAction = QAction(self.tr("Forward"), self)
        forwardAction.setShortcut(QKeySequence(self.tr("Right")))
        self.connect(forwardAction, SIGNAL("triggered()"), self.forward)
        self.addAction(forwardAction)
        
        backwardAction = QAction(self.tr("Backward"), self)
        backwardAction.setShortcut(QKeySequence(self.tr("Left")))
        self.connect(backwardAction, SIGNAL("triggered()"), self.backward)
        self.addAction(backwardAction)
        
        printAction = QAction(self.tr("Print"), self)
        printAction.setShortcut(QKeySequence(self.tr("Ctrl+P")))
        self.connect(printAction, SIGNAL("triggered()"), self.printPresentation)
        self.addAction(printAction)
    
    def loadPresentation(self):
    
        self.timeLine.stop()
        self.timeLine.setCurrentTime(0)
        
        self.slides = slides.init_slides(self.scene(), self)
        self.firstSlide = self.currentSlide = self.slides.slides[0]
        self.firstSlide.startAnimations()
    
    def forward(self):
    
        if self.timeLine.state() == QTimeLine.Running:
            if self.timeLine.direction() == QTimeLine.Forward:
                self.timeLine.setCurrentTime(self.transition_duration)
            else:
                self.timeLine.setDirection(QTimeLine.Forward)
        
        elif self.currentSlide.next:
        
            # Stop all animations in the slide then transit to the next slide.
            self.currentSlide.stopAnimations()
            self.timeLine.setDirection(QTimeLine.Forward)
            self.timeLine.start()
    
    def backward(self):
    
        if self.timeLine.state() == QTimeLine.Running:
            if self.timeLine.direction() == QTimeLine.Backward:
                self.timeLine.setCurrentTime(0)
            else:
                self.timeLine.setDirection(QTimeLine.Backward)
        
        elif self.currentSlide.previous:
        
            # Stop all animations in the slide then make the previous slide
            # the current slide and play its transition.
            self.currentSlide.stopAnimations()
            self.currentSlide = self.currentSlide.previous
            self.timeLine.setDirection(QTimeLine.Backward)
            self.timeLine.start()
    
    def updateTransition(self, value):
    
        self.currentSlide.transit(value, self)
    
    def updateSlide(self):
    
        if self.timeLine.direction() == QTimeLine.Forward:
            if self.currentSlide.next:
                self.currentSlide = self.currentSlide.next
        
        # Start animating the current slide.
        self.currentSlide.startAnimations()
    
    def printPresentation(self):
    
        self.currentSlide.stopAnimations()
        
        printer = QPrinter()
        printer.setPageSize(QPrinter.A4)
        printer.setOutputFormat(QPrinter.PdfFormat)
        printer.setPageOrder(QPrinter.FirstPageFirst)
        printer.setOrientation(QPrinter.Landscape)
        printer.setFontEmbeddingEnabled(True)
        printer.setOutputFileName("PyCon_Italia_2009_PyQt_for_Desktop_and_Embedded_Devices.pdf")
        
        rect = printer.pageRect()
        x, y = printer.paperRect().x(), printer.paperRect().y()
        
        painter = QPainter()
        painter.begin(printer)
        painter.setRenderHint(QPainter.SmoothPixmapTransform)
        painter.setRenderHint(QPainter.Antialiasing)
        
        qApp.setOverrideCursor(QCursor(Qt.WaitCursor))
        
        slide = self.firstSlide
        while slide:
        
            slide.transit(0, self)
            
            if slide.printable:
            
                painter.save()
                xscale = float(rect.width())/slide.rect().width()
                yscale = float(rect.height())/slide.rect().height()
                scale = min(xscale, yscale)
                
                painter.translate(x + rect.width()/2, y + rect.height()/2)
                painter.scale(scale, scale)
                painter.translate(-slide.rect().width()/2, -slide.rect().height()/2)
                
                self.render(painter, QRectF(0, 0, slides.Width, slides.Height))
                painter.restore()
                
                if slide.next:
                    printer.newPage()
            
            slide.transit(1, self)
            slide = slide.next
        
        painter.end()
        
        self.currentSlide.transit(0, self)
        self.currentSlide.startAnimations()
        qApp.restoreOverrideCursor()
    
    def wheelEvent(self, event):
        event.accept()
    
    def keyPressEvent(self, event):
    
        if event.key() in self.ignoreKeys:
            self.accept()
        else:
            return QGraphicsView.keyPressEvent(self, event)


if __name__ == "__main__":

    app = QApplication(sys.argv)
    
    scene = QGraphicsScene()
    view = View(scene)
    desktop = QDesktopWidget()
    view.setFixedSize(desktop.width(), desktop.height())
    view.loadPresentation()
    view.setFrameStyle(QFrame.NoFrame)
    view.setWindowTitle("PyQt for Desktop and Embedded Devices")

    view.showFullScreen()
    sys.exit(app.exec_())
